@@ -35,9 +35,6 @@
 #include <gnutls/dtls.h>
 #include "gnutls.h"
 
-#if GNUTLS_VERSION_NUMBER < 0x030200
-# define GNUTLS_DTLS1_2 202
-#endif
 #if GNUTLS_VERSION_NUMBER < 0x030400
 # define GNUTLS_CIPHER_CHACHA20_POLY1305 23
 #endif
@@ -121,7 +118,7 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 
 	buf_append(buf, "PSK-NEGOTIATE");
 
-	ret = gnutls_priority_init(&cache, vpninfo->gnutls_prio, NULL);
+	ret = gnutls_priority_init(&cache, vpninfo->ciphersuite_config, NULL);
 	if (ret < 0) {
 		buf->error = -EIO;
 		return;
@@ -178,9 +175,8 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
  * identifier in the client hello (draft-jay-tls-psk-identity-extension
  * could be used in the future). The session is not actually resumed.
  */
-static int start_dtls_psk_handshake(struct openconnect_info *vpninfo, int dtls_fd)
+static int start_dtls_psk_handshake(struct openconnect_info *vpninfo, gnutls_session_t dtls_ssl)
 {
-	gnutls_session_t dtls_ssl;
 	gnutls_datum_t key;
 	struct oc_text_buf *prio;
 	int err;
@@ -192,24 +188,14 @@ static int start_dtls_psk_handshake(struct openconnect_info *vpninfo, int dtls_f
 	}
 
 	prio = buf_alloc();
-	buf_append(prio, "%s:-VERS-TLS-ALL:+VERS-DTLS-ALL:-KX-ALL:+PSK", vpninfo->gnutls_prio);
+	buf_append(prio, "%s:-VERS-TLS-ALL:+VERS-DTLS-ALL:-KX-ALL:+PSK", vpninfo->ciphersuite_config);
 	if (buf_error(prio)) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to generate DTLS priority string\n"));
-		vpninfo->dtls_attempt_period = 0;
 		return buf_free(prio);
 	}
 
 
-	err = gnutls_init(&dtls_ssl, GNUTLS_CLIENT|GNUTLS_DATAGRAM|GNUTLS_NONBLOCK|GNUTLS_NO_EXTENSIONS);
-	if (err) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to initialize DTLS: %s\n"),
-			     gnutls_strerror(err));
-		goto fail;
-	}
-	gnutls_session_set_ptr(dtls_ssl, (void *) vpninfo);
-
 	err = gnutls_priority_set_direct(dtls_ssl, prio->data, NULL);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -230,9 +216,6 @@ static int start_dtls_psk_handshake(struct openconnect_info *vpninfo, int dtls_f
 		gnutls_session_set_id(dtls_ssl, &id);
 	}
 
-	gnutls_transport_set_ptr(dtls_ssl,
-				 (gnutls_transport_ptr_t)(intptr_t)dtls_fd);
-
 	/* set PSK credentials */
 	err = gnutls_psk_allocate_client_credentials(&vpninfo->psk_cred);
 	if (err < 0) {
@@ -278,27 +261,21 @@ static int start_dtls_psk_handshake(struct openconnect_info *vpninfo, int dtls_f
 	}
 
 	buf_free(prio);
-	vpninfo->dtls_ssl = dtls_ssl;
 	return 0;
+
  fail:
 	buf_free(prio);
-	gnutls_deinit(dtls_ssl);
 	gnutls_psk_free_client_credentials(vpninfo->psk_cred);
 	vpninfo->psk_cred = NULL;
-	vpninfo->dtls_attempt_period = 0;
 	return -EINVAL;
 }
 
-int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
+static int start_dtls_resume_handshake(struct openconnect_info *vpninfo, gnutls_session_t dtls_ssl)
 {
-	gnutls_session_t dtls_ssl;
 	gnutls_datum_t master_secret, session_id;
 	int err;
 	int cipher;
 
-	if (strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE") == 0)
-		return start_dtls_psk_handshake(vpninfo, dtls_fd);
-
 	for (cipher = 0; cipher < sizeof(gnutls_dtls_ciphers)/sizeof(gnutls_dtls_ciphers[0]); cipher++) {
 		if (gnutls_dtls_ciphers[cipher].cisco_dtls12 != vpninfo->cisco_dtls12 ||
 		    gnutls_check_version(gnutls_dtls_ciphers[cipher].min_gnutls_version) == NULL)
@@ -308,29 +285,19 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	}
 	vpn_progress(vpninfo, PRG_ERR, _("Unknown DTLS parameters for requested CipherSuite '%s'\n"),
 		     vpninfo->dtls_cipher);
-	vpninfo->dtls_attempt_period = 0;
-
 	return -EINVAL;
 
  found_cipher:
-	gnutls_init(&dtls_ssl, GNUTLS_CLIENT|GNUTLS_DATAGRAM|GNUTLS_NONBLOCK);
-	gnutls_session_set_ptr(dtls_ssl, (void *) vpninfo);
-
 	err = gnutls_priority_set_direct(dtls_ssl,
 					 gnutls_dtls_ciphers[cipher].prio,
 					 NULL);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to set DTLS priority: %s\n"),
-			     gnutls_strerror(err));
-		gnutls_deinit(dtls_ssl);
-		vpninfo->dtls_attempt_period = 0;
+			     _("Failed to set DTLS priority: '%s': %s\n"),
+			     gnutls_dtls_ciphers[cipher].prio, gnutls_strerror(err));
 		return -EINVAL;
 	}
 
-	gnutls_transport_set_ptr(dtls_ssl,
-				 (gnutls_transport_ptr_t)(intptr_t)dtls_fd);
-
 	gnutls_record_disable_padding(dtls_ssl);
 	master_secret.data = vpninfo->dtls_secret;
 	master_secret.size = sizeof(vpninfo->dtls_secret);
@@ -344,11 +311,88 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to set DTLS session parameters: %s\n"),
 			     gnutls_strerror(err));
-		gnutls_deinit(dtls_ssl);
-		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
 
+	return 0;
+}
+
+/*
+ * GnuTLS version between 3.6.3 and 3.6.12 send zero'ed ClientHello. Make sure
+ * we are not hitting that bug. Adapted from:
+ * https://gitlab.com/gnutls/gnutls/-/blob/3.6.13/tests/tls_hello_random_value.c
+ */
+static int check_client_hello_random(gnutls_session_t ttls_sess, unsigned int type,
+				     unsigned hook, unsigned int incoming, const gnutls_datum_t *msg)
+{
+	unsigned non_zero = 0, i;
+	struct openconnect_info *vpninfo = (struct openconnect_info *)gnutls_session_get_ptr(ttls_sess);
+
+	if (type == GNUTLS_HANDSHAKE_CLIENT_HELLO && hook == GNUTLS_HOOK_POST) {
+		gnutls_datum_t buf;
+		gnutls_session_get_random(ttls_sess, &buf, NULL);
+		if (buf.size != 32) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("GnuTLS used %d ClientHello random bytes; this should never happen\n"),
+				     buf.size);
+			return GNUTLS_E_INVALID_REQUEST;
+		}
+
+		for (i = 0; i < buf.size; ++i) {
+			if (buf.data[i] != 0) {
+				non_zero++;
+			}
+		}
+
+		/* The GnuTLS bug was that *all* bytes were zero, but as part of the unit test
+		 * they also slipped in a coincidental check on how well the random number
+		 * generator is behaving. Eight or more zeroes is a bad thing whatever the
+		 * reason for it. So we have the same check. */
+		if (non_zero <= 8) {
+			/* TODO: mention CVE number in log message once it's assigned */
+			vpn_progress(vpninfo, PRG_ERR,
+			     _("GnuTLS sent insecure ClientHello random. Upgrade to 3.6.13 or newer.\n"));
+			return GNUTLS_E_INVALID_REQUEST;
+		}
+	}
+
+	return 0;
+}
+
+int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
+{
+	gnutls_session_t dtls_ssl;
+	int err, ret;
+
+	err = gnutls_init(&dtls_ssl, GNUTLS_CLIENT|GNUTLS_DATAGRAM|GNUTLS_NONBLOCK|GNUTLS_NO_EXTENSIONS);
+	if (err) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to initialize DTLS: %s\n"),
+			     gnutls_strerror(err));
+		return -EINVAL;
+	}
+	gnutls_session_set_ptr(dtls_ssl, (void *) vpninfo);
+	gnutls_transport_set_ptr(dtls_ssl,
+				 (gnutls_transport_ptr_t)(intptr_t)dtls_fd);
+
+	if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE"))
+		ret = start_dtls_psk_handshake(vpninfo, dtls_ssl);
+	else
+		ret = start_dtls_resume_handshake(vpninfo, dtls_ssl);
+
+	if (ret) {
+		if (ret != -EAGAIN)
+			vpninfo->dtls_attempt_period = 0;
+		gnutls_deinit(dtls_ssl);
+		return ret;
+	}
+
+	if (gnutls_check_version_numeric(3,6,3) && !gnutls_check_version_numeric(3,6,13)) {
+		gnutls_handshake_set_hook_function(dtls_ssl, GNUTLS_HANDSHAKE_CLIENT_HELLO,
+						   GNUTLS_HOOK_POST, check_client_hello_random);
+	}
+
+
 	vpninfo->dtls_ssl = dtls_ssl;
 	return 0;
 }
