@@ -915,17 +915,20 @@ static int run_hip_script(struct openconnect_info *vpninfo)
 #endif
 
 	if (!vpninfo->csd_wrapper) {
-		vpn_progress(vpninfo, PRG_ERR,
-		             _("WARNING: Server asked us to submit HIP report with md5sum %s.\n"
-		               "    VPN connectivity may be disabled or limited without HIP report submission.\n    %s\n"),
-		             vpninfo->csd_token,
+		/* Only warn once */
+		if (!vpninfo->last_trojan) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("WARNING: Server asked us to submit HIP report with md5sum %s.\n"
+				       "    VPN connectivity may be disabled or limited without HIP report submission.\n    %s\n"),
+				     vpninfo->csd_token,
 #if defined(_WIN32) || defined(__native_client__)
-		             _("However, running the HIP report submission script on this platform is not yet implemented.")
+				     _("However, running the HIP report submission script on this platform is not yet implemented.")
 #else
-		             _("You need to provide a --csd-wrapper argument with the HIP report submission script.")
+				     _("You need to provide a --csd-wrapper argument with the HIP report submission script.")
 #endif
-			);
-		/* XXX: Many GlobalProtect VPNs work fine despite allegedly requiring HIP report submission */
+				);
+			/* XXX: Many GlobalProtect VPNs work fine despite allegedly requiring HIP report submission */
+		}
 		return 0;
 	}
 
@@ -981,7 +984,7 @@ static int run_hip_script(struct openconnect_info *vpninfo)
 		return ret;
 	} else {
 		/* in child: run HIP script */
-		char *hip_argv[32];
+		const char *hip_argv[32];
 		int i = 0;
 		close(pipefd[0]);
 		/* The duplicated fd does not have O_CLOEXEC */
@@ -991,20 +994,20 @@ static int run_hip_script(struct openconnect_info *vpninfo)
 			exit(1);
 
 		hip_argv[i++] = openconnect_utf8_to_legacy(vpninfo, vpninfo->csd_wrapper);
-		hip_argv[i++] = (char *)"--cookie";
+		hip_argv[i++] = "--cookie";
 		hip_argv[i++] = vpninfo->cookie;
 		if (vpninfo->ip_info.addr) {
-			hip_argv[i++] = (char *)"--client-ip";
-			hip_argv[i++] = (char *)vpninfo->ip_info.addr;
+			hip_argv[i++] = "--client-ip";
+			hip_argv[i++] = vpninfo->ip_info.addr;
 		}
 		if (vpninfo->ip_info.addr6) {
-			hip_argv[i++] = (char *)"--client-ipv6";
-			hip_argv[i++] = (char *)vpninfo->ip_info.addr6;
+			hip_argv[i++] = "--client-ipv6";
+			hip_argv[i++] = vpninfo->ip_info.addr6;
 		}
-		hip_argv[i++] = (char *)"--md5";
+		hip_argv[i++] = "--md5";
 		hip_argv[i++] = vpninfo->csd_token;
 		hip_argv[i++] = NULL;
-		execv(hip_argv[0], hip_argv);
+		execv(hip_argv[0], (char **)hip_argv);
 
 	out:
 		vpn_progress(vpninfo, PRG_ERR,
@@ -1044,21 +1047,22 @@ int gpst_setup(struct openconnect_info *vpninfo)
 	if (ret)
 		goto out;
 
-	/* Always check HIP once (even if no --csd-wrapper specified) */
-	if (!vpninfo->last_trojan) {
-		ret = check_and_maybe_submit_hip_report(vpninfo);
-		if (ret)
-			goto out;
-	}
+	/* Always check HIP after getting configuration */
+	ret = check_and_maybe_submit_hip_report(vpninfo);
+	if (ret)
+		goto out;
+
+        /* XX: last_trojan is used both as a sentinel to detect the
+         * first time we check/submit HIP, and for the mainloop to timeout
+         * when periodic re-checking is required.
+         */
+	vpninfo->last_trojan = time(NULL);
 
 	/* Default HIP re-checking to 3600 seconds unless already set by
-	 * --force-trojan or portal config. There's no point to rechecking
-	 * HIP if --csd-wrapper wasn't specified: either the VPN will
-	 * work despite lack of HIP submission, or it won't.
+	 * --force-trojan or portal config.
 	 */
-	if (vpninfo->csd_wrapper && !vpninfo->trojan_interval)
+	if (!vpninfo->trojan_interval)
 		vpninfo->trojan_interval = 3600;
-	vpninfo->last_trojan = time(NULL);
 
 	/* Connect tunnel immediately if ESP is not going to be used */
 	ret = gpst_connect(vpninfo);
