@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2014 Intel Corporation.
+ * Copyright © 2008-2015 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -54,19 +54,21 @@ static const gss_OID_desc gss_mech_spnego = {
 	(void *)&spnego_OID
 };
 
-static int gssapi_setup(struct openconnect_info *vpninfo, const char *service)
+static int gssapi_setup(struct openconnect_info *vpninfo, struct http_auth_state *auth_state,
+			const char *service, int proxy)
 {
 	OM_uint32 major, minor;
 	gss_buffer_desc token = GSS_C_EMPTY_BUFFER;
 	char *name;
 
-	if (asprintf(&name, "%s@%s", service, vpninfo->proxy) == -1)
+	if (asprintf(&name, "%s@%s", service,
+		     proxy ? vpninfo->proxy : vpninfo->hostname) == -1)
 		return -ENOMEM;
 	token.length = strlen(name);
 	token.value = name;
 
 	major = gss_import_name(&minor, &token, (gss_OID)GSS_C_NT_HOSTBASED_SERVICE,
-				&vpninfo->gss_target_name);
+				&auth_state->gss_target_name);
 	free(name);
 	if (GSS_ERROR(major)) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -80,77 +82,91 @@ static int gssapi_setup(struct openconnect_info *vpninfo, const char *service)
 #define GSSAPI_CONTINUE	2
 #define GSSAPI_COMPLETE	3
 
-int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *hdrbuf)
+int gssapi_authorization(struct openconnect_info *vpninfo, int proxy,
+			 struct http_auth_state *auth_state,
+			 struct oc_text_buf *hdrbuf)
 {
 	OM_uint32 major, minor;
 	gss_buffer_desc in = GSS_C_EMPTY_BUFFER;
 	gss_buffer_desc out = GSS_C_EMPTY_BUFFER;
 	gss_OID mech = GSS_C_NO_OID;
 
-	if (vpninfo->auth[AUTH_TYPE_GSSAPI].state == AUTH_AVAILABLE && gssapi_setup(vpninfo, "HTTP")) {
-		vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_FAILED;
+	if (auth_state->state == AUTH_AVAILABLE && gssapi_setup(vpninfo, auth_state, "HTTP", proxy)) {
+		auth_state->state = AUTH_FAILED;
 		return -EIO;
 	}
 
-	if (vpninfo->auth[AUTH_TYPE_GSSAPI].challenge && *vpninfo->auth[AUTH_TYPE_GSSAPI].challenge) {
+	if (auth_state->challenge && *auth_state->challenge) {
 		int len = -EINVAL;
-		in.value = openconnect_base64_decode(&len, vpninfo->auth[AUTH_TYPE_GSSAPI].challenge);
+		in.value = openconnect_base64_decode(&len, auth_state->challenge);
 		if (!in.value)
 			return len;
 		in.length = len;
-	} else if (vpninfo->auth[AUTH_TYPE_GSSAPI].state > AUTH_AVAILABLE) {
+	} else if (auth_state->state > AUTH_AVAILABLE) {
 		/* This indicates failure. We were trying, but got an empty
 		   'Proxy-Authorization: Negotiate' header back from the server
 		   implying that we should start again... */
 		goto fail_gssapi;
 	}
 
-	major = gss_init_sec_context(&minor, GSS_C_NO_CREDENTIAL, &vpninfo->gss_context,
-				     vpninfo->gss_target_name, (gss_OID)&gss_mech_spnego, GSS_C_MUTUAL_FLAG,
-				     GSS_C_INDEFINITE, GSS_C_NO_CHANNEL_BINDINGS, &in, &mech,
-				     &out, NULL, NULL);
+	major = gss_init_sec_context(&minor, GSS_C_NO_CREDENTIAL,
+				     &auth_state->gss_context,
+				     auth_state->gss_target_name,
+				     (gss_OID)&gss_mech_spnego,
+				     GSS_C_MUTUAL_FLAG, GSS_C_INDEFINITE,
+				     GSS_C_NO_CHANNEL_BINDINGS, &in,
+				     &mech, &out, NULL, NULL);
 	if (in.value)
 		free(in.value);
 
 	if (major == GSS_S_COMPLETE)
-		vpninfo->auth[AUTH_TYPE_GSSAPI].state = GSSAPI_COMPLETE;
+		auth_state->state = GSSAPI_COMPLETE;
 	else if (major == GSS_S_CONTINUE_NEEDED)
-		vpninfo->auth[AUTH_TYPE_GSSAPI].state = GSSAPI_CONTINUE;
+		auth_state->state = GSSAPI_CONTINUE;
 	else {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error generating GSSAPI response:\n"));
 		print_gss_err(vpninfo, "gss_init_sec_context()", mech, major, minor);
 	fail_gssapi:
-		vpninfo->auth[AUTH_TYPE_GSSAPI].state = AUTH_FAILED;
-		cleanup_gssapi_auth(vpninfo);
+		auth_state->state = AUTH_FAILED;
+		cleanup_gssapi_auth(vpninfo, auth_state);
 		/* If we were *trying*, then -EAGAIN. Else -ENOENT to let another
 		   auth method try without having to reconnect first. */
 		return in.value ? -EAGAIN : -ENOENT;
 	}
-	buf_append(hdrbuf, "Proxy-Authorization: Negotiate ");
+	buf_append(hdrbuf, "%sAuthorization: Negotiate ", proxy ? "Proxy-" : "");
 	buf_append_base64(hdrbuf, out.value, out.length);
 	buf_append(hdrbuf, "\r\n");
 
 	gss_release_buffer(&minor, &out);
-	if (!vpninfo->auth[AUTH_TYPE_GSSAPI].challenge)
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("Attempting GSSAPI authentication to proxy\n"));
+	if (!auth_state->challenge) {
+		if (proxy)
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Attempting GSSAPI authentication to proxy\n"));
+		else
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Attempting GSSAPI authentication to server '%s'\n"),
+				     vpninfo->hostname);
+	}
+
 	return 0;
 }
 
-void cleanup_gssapi_auth(struct openconnect_info *vpninfo)
+/* auth_state is NULL when called from socks_gssapi_auth() */
+void cleanup_gssapi_auth(struct openconnect_info *vpninfo,
+			 struct http_auth_state *auth_state)
 {
 	OM_uint32 minor;
 
-	if (vpninfo->gss_target_name != GSS_C_NO_NAME)
-		gss_release_name(&minor, &vpninfo->gss_target_name);
+	if (auth_state->gss_target_name != GSS_C_NO_NAME)
+		gss_release_name(&minor, &auth_state->gss_target_name);
 
-	if (vpninfo->gss_context != GSS_C_NO_CONTEXT)
-		gss_delete_sec_context(&minor, &vpninfo->gss_context, GSS_C_NO_BUFFER);
+	if (auth_state->gss_context != GSS_C_NO_CONTEXT)
+		gss_delete_sec_context(&minor, &auth_state->gss_context, GSS_C_NO_BUFFER);
 
 	/* Shouldn't be necessary, but make sure... */
-	vpninfo->gss_target_name = GSS_C_NO_NAME;
-	vpninfo->gss_context = GSS_C_NO_CONTEXT;
+	auth_state->gss_target_name = GSS_C_NO_NAME;
+	auth_state->gss_context = GSS_C_NO_CONTEXT;
 }
 
 int socks_gssapi_auth(struct openconnect_info *vpninfo)
@@ -162,16 +178,17 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 	unsigned char *pktbuf;
 	int i;
 	int ret = -EIO;
+	struct http_auth_state *auth_state = &vpninfo->proxy_auth[AUTH_TYPE_GSSAPI];
 
-	if (gssapi_setup(vpninfo, "rcmd"))
+	if (gssapi_setup(vpninfo, auth_state, "rcmd", 1))
 		return -EIO;
 
 	pktbuf = malloc(65538);
 	if (!pktbuf)
 		return -ENOMEM;
 	while (1) {
-		major = gss_init_sec_context(&minor, GSS_C_NO_CREDENTIAL, &vpninfo->gss_context,
-					     vpninfo->gss_target_name, (gss_OID)&gss_mech_spnego,
+		major = gss_init_sec_context(&minor, GSS_C_NO_CREDENTIAL, &auth_state->gss_context,
+					     auth_state->gss_target_name, (gss_OID)&gss_mech_spnego,
 					     GSS_C_MUTUAL_FLAG | GSS_C_REPLAY_FLAG | GSS_C_DELEG_FLAG | GSS_C_SEQUENCE_FLAG,
 					     GSS_C_INDEFINITE, GSS_C_NO_CHANNEL_BINDINGS, &in, &mech,
 					     &out, NULL, NULL);
@@ -198,8 +215,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 
 		pktbuf[0] = 1; /* ver */
 		pktbuf[1] = 1; /* mtyp */
-		pktbuf[2] = (out.length >> 8) & 0xff;
-		pktbuf[3] = out.length & 0xff;
+		store_be16(pktbuf + 2, out.length);
 		memcpy(pktbuf + 4, out.value, out.length);
 
 		free(out.value);
@@ -232,7 +248,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 				     pktbuf[1]);
 			break;
 		}
-		in.length = (pktbuf[2] << 8) | pktbuf[3];
+		in.length = load_be16(pktbuf + 2);
 		in.value = pktbuf;
 
 		if (!in.length) {
@@ -260,7 +276,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 		in.value = pktbuf;
 		in.length = 1;
 
-		major = gss_wrap(&minor, vpninfo->gss_context, 0,
+		major = gss_wrap(&minor, auth_state->gss_context, 0,
 				 GSS_C_QOP_DEFAULT, &in, NULL, &out);
 		if (major != GSS_S_COMPLETE) {
 			print_gss_err(vpninfo, "gss_wrap()", mech, major, minor);
@@ -269,8 +285,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 
 		pktbuf[0] = 1;
 		pktbuf[1] = 2;
-		pktbuf[2] = (out.length >> 8) & 0xff;
-		pktbuf[3] = out.length & 0xff;
+		store_be16(pktbuf + 2, out.length);
 		memcpy(pktbuf + 4, out.value, out.length);
 
 		free(out.value);
@@ -293,7 +308,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 				     strerror(-i));
 			goto err;
 		}
-		in.length = (pktbuf[2] << 8) | pktbuf[3];
+		in.length = load_be16(pktbuf + 2);
 		in.value = pktbuf;
 
 		i = vpninfo->ssl_read(vpninfo, (void *)pktbuf, in.length);
@@ -307,7 +322,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 			     _("Got GSSAPI protection response of %zu bytes: %02x %02x %02x %02x\n"),
 			     in.length, pktbuf[0], pktbuf[1], pktbuf[2], pktbuf[3]);
 
-		major = gss_unwrap(&minor, vpninfo->gss_context, &in, &out, NULL, GSS_C_QOP_DEFAULT);
+		major = gss_unwrap(&minor, auth_state->gss_context, &in, &out, NULL, GSS_C_QOP_DEFAULT);
 		if (major != GSS_S_COMPLETE) {
 			print_gss_err(vpninfo, "gss_unwrap()", mech, major, minor);
 			goto err;
@@ -338,7 +353,7 @@ int socks_gssapi_auth(struct openconnect_info *vpninfo)
 		ret = 0;
 	}
  err:
-	cleanup_gssapi_auth(vpninfo);
+	cleanup_gssapi_auth(vpninfo, NULL);
 	free(pktbuf);
 
 	return ret;
