@@ -195,7 +195,13 @@ int openconnect_get_supported_protocols(struct oc_vpn_proto **protos)
 	struct oc_vpn_proto *pr;
 	int i;
 
-	*protos = pr = calloc(NR_PROTOS, sizeof(*pr));
+        /* The original version of this function included an all-zero
+         * sentinel value at the end of the array, so we must continue
+         * to do so for ABI compatibility even though it's
+         * functionally redundant as a marker of the array's length,
+         * along with the explicit length in the return value.
+         */
+	*protos = pr = calloc(NR_PROTOS + 1, sizeof(*pr));
 	if (!pr)
 		return -ENOMEM;
 
