@@ -659,10 +659,12 @@ static void print_supported_protocols(void)
 {
 	const char *comma = ", ", *sep = comma + 1;
 	struct oc_vpn_proto *protos, *p;
+	int n;
 
-	if (openconnect_get_supported_protocols(&protos)>=0) {
+	n = openconnect_get_supported_protocols(&protos);
+        if (n>=0) {
 		printf(_("Supported protocols:"));
-		for (p=protos; p->name; p++) {
+		for (p=protos; n; p++, n--) {
 			printf("%s%s%s", sep, p->name, p==protos ? _(" (default)") : "");
 			sep = comma;
 		}
@@ -674,10 +676,12 @@ static void print_supported_protocols(void)
 static void print_supported_protocols_usage(void)
 {
 	struct oc_vpn_proto *protos, *p;
+	int n;
 
-	if (openconnect_get_supported_protocols(&protos)>=0) {
+	n = openconnect_get_supported_protocols(&protos);
+        if (n>=0) {
 		printf("\n%s:\n", _("Set VPN protocol"));
-		for (p=protos; p->name; p++)
+		for (p=protos; n; p++, n--)
 			printf("      --protocol=%-16s %s%s\n",
 				   p->name, p->description, p==protos ? _(" (default)") : "");
 		openconnect_free_supported_protocols(protos);
@@ -1100,6 +1104,278 @@ static void get_uids(const char *config_arg, uid_t *uid, gid_t *gid)
 }
 #endif
 
+static int complete_words(const char *comp_opt, int prefixlen, ...)
+{
+	int partlen = strlen(comp_opt + prefixlen);
+	va_list vl;
+	char *check;
+
+	va_start(vl, prefixlen);
+	while ( (check = va_arg(vl, char *)) ) {
+		if (!strncmp(comp_opt + prefixlen, check, partlen))
+			printf("%.*s%s\n", prefixlen, comp_opt, check);
+	}
+	va_end(vl);
+	return 0;
+}
+
+static int autocomplete_special(const char *verb, const char *prefix,
+				int prefixlen, const char *filterpat)
+{
+	printf("%s\n", verb);
+	printf("%s\n", filterpat ? : "''");
+
+	if (prefixlen)
+		printf("%.*s\n", prefixlen, prefix);
+	return 0;
+}
+
+static int autocomplete(int argc, char **argv)
+{
+	int opt;
+	const char *comp_cword = getenv("COMP_CWORD");
+	char *comp_opt;
+	int cword, longidx, prefixlen = 0;
+
+	/* Skip over the --autocomplete */
+	argc--;
+	argv++;
+
+	if (!comp_cword)
+		return -EINVAL;
+
+	cword = atoi(comp_cword);
+	if (cword <= 0 || cword > argc)
+		return -EINVAL;
+
+	comp_opt = argv[cword];
+	if (!comp_opt)
+		return -EINVAL;
+
+	opterr = 0;
+
+	while (argv[optind]) {
+		/* If optind is the one that is being autocompleted, don't
+		 * let getopt_long() see it; we process it directly. */
+		if (argv[optind] == comp_opt) {
+			if (!strncmp(comp_opt, "--", 2)) {
+				const char *arg = strchr(comp_opt, '=');
+				int matchlen;
+
+				if (arg) {
+					/* We have --option=... so complete the arg */
+					matchlen = arg - comp_opt - 2;
+					for (longidx = 0; long_options[longidx].name; longidx++) {
+						if (!strncmp(comp_opt + 2, long_options[longidx].name, matchlen)) {
+							prefixlen = matchlen + 3;
+							opt = long_options[longidx].val;
+							goto got_opt;
+						}
+					}
+				} else {
+					/* Not --option= just --opt so complete the option name(s) */
+					comp_opt += 2;
+				autocomplete_optname:
+					matchlen = strlen(comp_opt);
+					for (longidx = 0; long_options[longidx].name; longidx++) {
+						if (!strncmp(comp_opt, long_options[longidx].name, matchlen)) {
+							printf("--%s\n", long_options[longidx].name);
+						}
+					}
+				}
+			} else if (comp_opt[0] == '-') {
+				if (!comp_opt[1]) {
+					/* Just a single dash. Autocomplete like '--' with all the (long) options */
+					comp_opt++;
+					goto autocomplete_optname;
+				}
+				/* Single-char -X option, with or without an argument. */
+				for (longidx = 0; long_options[longidx].name; longidx++) {
+					if (comp_opt[1] == long_options[longidx].val) {
+						if (comp_opt[2]) {
+							if (long_options[longidx].has_arg) {
+								prefixlen = 2;
+								opt = long_options[longidx].val;
+								goto got_opt;
+							}
+						} else {
+							/* Just the option; complete to the long name of same. */
+							printf("--%s\n", long_options[longidx].name);
+						}
+						break;
+					}
+				}
+			} else
+				printf("HOSTNAME\n");
+
+			return 0;
+		}
+
+		/* Skip over non-option elements, in an attempt to prevent
+		 * getopt_long() from reordering the array as we go. The problem
+		 * is that we've seen it *delay* the reordering. So it processes
+		 * the argv element *after* the non-option, but argv[optind] is
+		 * still pointing to the non-option. */
+		if (argv[optind][0] != '-') {
+			optind++;
+			continue;
+		}
+
+		opt = getopt_long(argc, argv,
+#ifdef _WIN32
+				  "C:c:Dde:F:g:hi:k:m:P:p:Q:qs:u:Vvx:",
+#else
+				  "bC:c:Dde:F:g:hi:k:lm:P:p:Q:qSs:U:u:Vvx:",
+#endif
+				  long_options, &longidx);
+
+		if (opt == -1)
+			break;
+
+		if (optarg == comp_opt) {
+			prefixlen = 0;
+		got_opt:
+			switch (opt) {
+			case 'k': /* --sslkey */
+			case 'c': /* --certificate */
+				if (!strncmp(comp_opt + prefixlen, "pkcs11:", 7)) {
+					/* We could do clever things here... */
+					return 0; /* .. but we don't. */
+				}
+				autocomplete_special("FILENAME", comp_opt, prefixlen, "!*.@(pem|der|p12|crt)");
+				break;
+
+			case OPT_CAFILE: /* --cafile */
+				autocomplete_special("FILENAME", comp_opt, prefixlen, "!*.@(pem|der|crt)");
+				break;
+
+			case 'x': /* --xmlconfig */
+				autocomplete_special("FILENAME", comp_opt, prefixlen, "!*.xml");
+				break;
+
+			case OPT_CONFIGFILE: /* --config */
+			case OPT_PIDFILE: /* --pid-file */
+				autocomplete_special("FILENAME", comp_opt, prefixlen, NULL);
+				break;
+
+			case 's': /* --script */
+			case OPT_CSD_WRAPPER: /* --csd-wrapper */
+				autocomplete_special("EXECUTABLE", comp_opt, prefixlen, NULL);
+				break;
+
+			case OPT_LOCAL_HOSTNAME: /* --local-hostname */
+				autocomplete_special("HOSTNAME", comp_opt, prefixlen, NULL);
+				break;
+
+			case OPT_CSD_USER: /* --csd-user */
+			case 'U': /* --setuid */
+				autocomplete_special("USERNAME", comp_opt, prefixlen, NULL);
+				break;
+
+			case OPT_OS: /* --os */
+				complete_words(comp_opt, prefixlen, "mac-intel", "android",
+					       "linux-64", "linux", "apple-ios",
+					       "win", NULL);
+				break;
+
+			case OPT_COMPRESSION: /* --compression */
+				complete_words(comp_opt, prefixlen, "none", "off", "all",
+					       "stateless", NULL);
+				break;
+
+			case OPT_PROTOCOL: /* --protocol */
+			{
+				struct oc_vpn_proto *protos, *p;
+				int partlen = strlen(comp_opt + prefixlen);
+
+				if (openconnect_get_supported_protocols(&protos) >= 0) {
+					for (p = protos; p->name; p++) {
+						if(!strncmp(comp_opt + prefixlen, p->name, partlen))
+							printf("%.*s%s\n", prefixlen, comp_opt, p->name);
+					}
+					free(protos);
+				}
+				break;
+			}
+
+			case OPT_HTTP_AUTH: /* --http-auth */
+			case OPT_PROXY_AUTH: /* --proxy-auth */
+				/* FIXME: Expand latest list item */
+				break;
+
+			case OPT_TOKEN_MODE: /* --token-mode */
+				complete_words(comp_opt, prefixlen, "totp", "hotp", "oidc", NULL);
+				if (openconnect_has_stoken_support())
+					complete_words(comp_opt, prefixlen, "rsa", NULL);
+				if (openconnect_has_yubioath_support())
+					complete_words(comp_opt, prefixlen, "yubioath", NULL);
+				break;
+
+			case OPT_TOKEN_SECRET: /* --token-secret */
+				switch (comp_opt[prefixlen]) {
+				case '@':
+					prefixlen++;
+					/* Fall through */
+				case 0:
+				case '/':
+					autocomplete_special("FILENAME", comp_opt, prefixlen, NULL);
+					break;
+				}
+				break;
+
+			case 'i': /* --interface */
+				/* FIXME: Enumerate available tun devices */
+				break;
+
+			case OPT_SERVERCERT: /* --servercert */
+				/* We could do something really evil here and actually
+				 * connect, then return the result? */
+				break;
+
+			/* No autocmplete for these but handle them explicitly so that
+			 * we can have automatic checking for *accidentally* unhandled
+			 * options. Right after we do automated checking of man page
+			 * entries and --help output for all supported options too. */
+
+			case 'e': /* --cert-expire-warning */
+			case 'C': /* --cookie */
+			case 'g': /* --usergroup */
+			case 'm': /* --mtu */
+			case OPT_BASEMTU: /* --base-mtu */
+			case 'p': /* --key-password */
+			case 'P': /* --proxy */
+			case 'u': /* --user */
+			case 'Q': /* --queue-len */
+			case OPT_RECONNECT_TIMEOUT: /* --reconnect-timeout */
+			case OPT_AUTHGROUP: /* --authgroup */
+			case OPT_RESOLVE: /* --resolve */
+			case OPT_USERAGENT: /* --useragent */
+			case OPT_VERSION: /* --version-string */
+			case OPT_FORCE_DPD: /* --force-dpd */
+			case OPT_FORCE_TROJAN: /* --force-trojan */
+			case OPT_DTLS_LOCAL_PORT: /* --dtls-local-port */
+			case 'F': /* --form-entry */
+			case OPT_GNUTLS_DEBUG: /* --gnutls-debug */
+			case OPT_CIPHERSUITES: /* --gnutls-priority */
+			case OPT_DTLS_CIPHERS: /* --dtls-ciphers */
+			case OPT_DTLS12_CIPHERS: /* --dtls12-ciphers */
+				break;
+
+			default:
+				fprintf(stderr, _("Unhandled autocomplete for option %d '--%s'. Please report.\n"),
+					opt, long_options[longidx].name);
+				return -ENOENT;
+			}
+
+			return 0;
+		}
+	}
+
+	/* Ths only non-option argument we accept as a hostname */
+	printf("HOSTNAME\n");
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	struct openconnect_info *vpninfo;
@@ -1137,6 +1413,9 @@ int main(int argc, char **argv)
 		fprintf(stderr,
 			_("WARNING: Cannot set locale: %s\n"), strerror(errno));
 
+	if (argc > 2 && !strcmp(argv[1], "--autocomplete"))
+		return autocomplete(argc, argv);
+
 #ifdef HAVE_NL_LANGINFO
 	charset = nl_langinfo(CODESET);
 	if (charset && strcmp(charset, "UTF-8"))
