@@ -57,7 +57,7 @@ void destroy_esp_ciphers(struct esp *esp)
 	}
 }
 
-static int init_esp_ciphers(struct openconnect_info *vpninfo, struct esp *esp,
+static int init_esp_cipher(struct openconnect_info *vpninfo, struct esp *esp,
 			    const EVP_MD *macalg, const EVP_CIPHER *encalg, int decrypt)
 {
 	int ret;
@@ -78,8 +78,7 @@ static int init_esp_ciphers(struct openconnect_info *vpninfo, struct esp *esp,
 	if (decrypt)
 		ret = EVP_DecryptInit_ex(esp->cipher, encalg, NULL, esp->enc_key, NULL);
 	else {
-		ret = RAND_bytes((void *)&esp->iv, sizeof(esp->iv)) &&
-			EVP_EncryptInit_ex(esp->cipher, encalg, NULL, esp->enc_key, esp->iv);
+		ret = EVP_EncryptInit_ex(esp->cipher, encalg, NULL, esp->enc_key, esp->iv);
 	}
 
 	if (!ret) {
@@ -103,28 +102,21 @@ static int init_esp_ciphers(struct openconnect_info *vpninfo, struct esp *esp,
 		openconnect_report_ssl_errors(vpninfo);
 		destroy_esp_ciphers(esp);
 	}
-	esp->seq = 0;
-	esp->seq_backlog = 0;
+
 	return 0;
 }
 
-int setup_esp_keys(struct openconnect_info *vpninfo, int new_keys)
+int init_esp_ciphers(struct openconnect_info *vpninfo, struct esp *esp_out, struct esp *esp_in)
 {
-	struct esp *esp_in;
 	const EVP_CIPHER *encalg;
 	const EVP_MD *macalg;
 	int ret;
 
-	if (vpninfo->dtls_state == DTLS_DISABLED)
-		return -EOPNOTSUPP;
-	if (!vpninfo->dtls_addr)
-		return -EINVAL;
-
 	switch (vpninfo->esp_enc) {
-	case 0x02:
+	case ENC_AES_128_CBC:
 		encalg = EVP_aes_128_cbc();
 		break;
-	case 0x05:
+	case ENC_AES_256_CBC:
 		encalg = EVP_aes_256_cbc();
 		break;
 	default:
@@ -132,54 +124,36 @@ int setup_esp_keys(struct openconnect_info *vpninfo, int new_keys)
 	}
 
 	switch (vpninfo->esp_hmac) {
-	case 0x01:
+	case HMAC_MD5:
 		macalg = EVP_md5();
 		break;
-	case 0x02:
+	case HMAC_SHA1:
 		macalg = EVP_sha1();
 		break;
+	case HMAC_SHA256:
+		macalg = EVP_sha256();
+		break;
 	default:
 		return -EINVAL;
 	}
 
-	if (new_keys) {
-		vpninfo->old_esp_maxseq = vpninfo->esp_in[vpninfo->current_esp_in].seq + 32;
-		vpninfo->current_esp_in ^= 1;
-	}
-
-	esp_in = &vpninfo->esp_in[vpninfo->current_esp_in];
-
-	if (new_keys) {
-		if (!RAND_bytes((void *)&esp_in->spi, sizeof(esp_in->spi)) ||
-		    !RAND_bytes((void *)&esp_in->enc_key, vpninfo->enc_key_len) ||
-		    !RAND_bytes((void *)&esp_in->hmac_key, vpninfo->hmac_key_len) ) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Failed to generate random keys for ESP:\n"));
-			openconnect_report_ssl_errors(vpninfo);
-			return -EIO;
-		}
-	}
-
-	ret = init_esp_ciphers(vpninfo, &vpninfo->esp_out, macalg, encalg, 0);
+	ret = init_esp_cipher(vpninfo, &vpninfo->esp_out, macalg, encalg, 0);
 	if (ret)
 		return ret;
 
-	ret = init_esp_ciphers(vpninfo, esp_in, macalg, encalg, 1);
+	ret = init_esp_cipher(vpninfo, esp_in, macalg, encalg, 1);
 	if (ret) {
 		destroy_esp_ciphers(&vpninfo->esp_out);
 		return ret;
 	}
 
-	if (vpninfo->dtls_state == DTLS_NOSECRET)
-		vpninfo->dtls_state = DTLS_SECRET;
-	vpninfo->pkt_trailer = 16 + 20; /* 16 for pad, 20 for HMAC (of which we use 16) */
 	return 0;
 }
 
 /* pkt->len shall be the *payload* length. Omitting the header and the 12-byte HMAC */
 int decrypt_esp_packet(struct openconnect_info *vpninfo, struct esp *esp, struct pkt *pkt)
 {
-	unsigned char hmac_buf[20];
+	unsigned char hmac_buf[MAX_HMAC_SIZE];
 	unsigned int hmac_len = sizeof(hmac_buf);
 	int crypt_len = pkt->len;
 
@@ -187,7 +161,7 @@ int decrypt_esp_packet(struct openconnect_info *vpninfo, struct esp *esp, struct
 	HMAC_Update(esp->hmac, (void *)&pkt->esp, sizeof(pkt->esp) + pkt->len);
 	HMAC_Final(esp->hmac, hmac_buf, &hmac_len);
 
-	if (memcmp(hmac_buf, pkt->data + pkt->len, 12)) {
+	if (memcmp(hmac_buf, pkt->data + pkt->len, vpninfo->hmac_out_len)) {
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Received ESP packet with invalid HMAC\n"));
 		return -EINVAL;
@@ -215,26 +189,11 @@ int decrypt_esp_packet(struct openconnect_info *vpninfo, struct esp *esp, struct
 	return 0;
 }
 
-int encrypt_esp_packet(struct openconnect_info *vpninfo, struct pkt *pkt)
+int encrypt_esp_packet(struct openconnect_info *vpninfo, struct pkt *pkt, int crypt_len)
 {
-	int i, padlen;
 	int blksize = 16;
-	unsigned int hmac_len = 20;
-	int crypt_len;
+	unsigned int hmac_len = vpninfo->hmac_out_len;
 
-	/* This gets much more fun if the IV is variable-length */
-	pkt->esp.spi = vpninfo->esp_out.spi;
-	pkt->esp.seq = htonl(vpninfo->esp_out.seq++);
-
-	padlen = blksize - 1 - ((pkt->len + 1) % blksize);
-	for (i=0; i<padlen; i++)
-		pkt->data[pkt->len + i] = i + 1;
-	pkt->data[pkt->len + padlen] = padlen;
-	pkt->data[pkt->len + padlen + 1] = 0x04; /* Legacy IP */
-
-	memcpy(pkt->esp.iv, vpninfo->esp_out.iv, 16);
-
-	crypt_len = pkt->len + padlen + 2;
 	if (!EVP_EncryptUpdate(vpninfo->esp_out.cipher, pkt->data, &crypt_len,
 			       pkt->data, crypt_len)) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -249,5 +208,5 @@ int encrypt_esp_packet(struct openconnect_info *vpninfo, struct pkt *pkt)
 
 	EVP_EncryptUpdate(vpninfo->esp_out.cipher, vpninfo->esp_out.iv, &blksize,
 			  pkt->data + crypt_len + hmac_len - blksize, blksize);
-	return sizeof(pkt->esp) + crypt_len + 12;
+	return 0;
 }
